package net.jpountz.xxhash;

/*
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

enum XXHashConstants {
  ;

  static final int PRIME1 = -1640531535;
  static final int PRIME2 = -2048144777;
  static final int PRIME3 = -1028477379;
  static final int PRIME4 = 668265263;
  static final int PRIME5 = 374761393;

  static final long PRIME64_1 = -7046029288634856825L; //11400714785074694791
  static final long PRIME64_2 = -4417276706812531889L; //14029467366897019727
  static final long PRIME64_3 = 1609587929392839161L;
  static final long PRIME64_4 = -8796714831421723037L; //9650029242287828579
  static final long PRIME64_5 = 2870177450012600261L;
}
